import torch
import torch.nn as nn
from scvi.distributions import NegativeBinomial
from scvi.nn import one_hot

from ._utils import ResnetFC


class ExpActivation(nn.Module):
    def __init__(self) -> None:
        super().__init__()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return torch.exp(x)


class DecoderZX(nn.Module):
    """Parameterizes the counts likelihood for the data given the latent variables."""

    def __init__(
        self,
        n_in,
        n_out,
        n_nuisance,
        linear_decoder,
        n_hidden=128,
        activation="softmax",
    ):
        super().__init__()
        if activation == "softmax":
            activation_ = nn.Softmax(-1)
        elif activation == "softplus":
            activation_ = nn.Softplus()
        elif activation == "exp":
            activation_ = ExpActivation()
        elif activation == "sigmoid":
            activation_ = nn.Sigmoid()
        else:
            raise ValueError("activation must be one of 'softmax' or 'softplus'")
        self.linear_decoder = linear_decoder
        self.n_nuisance = n_nuisance
        self.n_latent = n_in - n_nuisance
        if linear_decoder:
            self.amat = nn.Linear(self.n_latent, n_out, bias=False)
            self.amat_site = nn.Parameter(
                torch.randn(self.n_nuisance, self.n_latent, n_out)
            )
            self.offsets = nn.Parameter(torch.randn(self.n_nuisance, n_out))
            self.dropout_ = nn.Dropout(0.1)
            self.activation_ = activation_

        else:
            self.px_mean = ResnetFC(
                n_in=n_in,
                n_out=n_out,
                n_hidden=n_hidden,
                activation=activation_,
            )
        self.px_r = nn.Parameter(torch.randn(n_out))

    def forward(self, z, size_factor):
        if self.linear_decoder:
            nuisance_oh = z[..., -self.n_nuisance :]
            z0 = z[..., : -self.n_nuisance]
            x1 = self.amat(z0)

            nuisance_ids = torch.argmax(nuisance_oh, -1)
            As = self.amat_site[nuisance_ids]
            z0_detach = self.dropout_(z0.detach())[..., None]
            x2 = (As * z0_detach).sum(-2)
            offsets = self.offsets[nuisance_ids]
            mu = x1 + x2 + offsets
            mu = self.activation_(mu)
        else:
            mu = self.px_mean(z)
        mu = mu * size_factor
        return NegativeBinomial(mu=mu, theta=self.px_r.exp())


class LinearDecoderUZ(nn.Module):
    def __init__(
        self,
        n_latent,
        n_sample,
        n_out,
        scaler=False,
        scaler_n_hidden=32,
    ):
        super().__init__()
        self.n_latent = n_latent
        self.n_sample = n_sample
        self.n_out = n_out

        self.amat_sample = nn.Parameter(torch.randn(n_sample, self.n_latent, n_out))
        self.offsets = nn.Parameter(torch.randn(n_sample, n_out))

        self.scaler = None
        if scaler:
            self.scaler = nn.Sequential(
                nn.Linear(n_latent + n_sample, scaler_n_hidden),
                nn.LayerNorm(scaler_n_hidden),
                nn.ReLU(),
                nn.Linear(scaler_n_hidden, 1),
                nn.Sigmoid(),
            )

    def forward(self, u, sample_id):
        sample_id_ = sample_id.long().squeeze()
        As = self.amat_sample[sample_id_]

        u_detach = u.detach()[..., None]
        z2 = (As * u_detach).sum(-2)
        offsets = self.offsets[sample_id_]
        delta = z2 + offsets
        if self.scaler is not None:
            sample_oh = one_hot(sample_id, self.n_sample)
            if u.ndim != sample_oh.ndim:
                sample_oh = sample_oh[None].expand(u.shape[0], *sample_oh.shape)
            inputs = torch.cat([u.detach(), sample_oh], -1)
            delta = delta * self.scaler(inputs)
        return u + delta


class DecoderUZ(nn.Module):
    def __init__(
        self,
        n_latent,
        n_latent_sample,
        n_out,
        dropout_rate=0.0,
        n_layers=1,
        n_hidden=128,
    ):
        super().__init__()
        self.n_latent = n_latent
        self.n_latent_sample = n_latent_sample
        self.n_in = n_latent + n_latent_sample
        self.n_out = n_out

        arch_mod = self.construct_arch(self.n_in, n_hidden, n_layers, dropout_rate) + [
            nn.Linear(n_hidden, self.n_out, bias=False)
        ]
        self.mod = nn.Sequential(*arch_mod)

        arch_scaler = self.construct_arch(
            self.n_latent, n_hidden, n_layers, dropout_rate
        ) + [nn.Linear(n_hidden, 1)]
        self.scaler = nn.Sequential(*arch_scaler)
        self.scaler.append(nn.Sigmoid())

    @staticmethod
    def construct_arch(n_inputs, n_hidden, n_layers, dropout_rate):
        """Initializes MLP architecture"""

        block_inputs = [
            nn.Linear(n_inputs, n_hidden),
            nn.BatchNorm1d(n_hidden),
            nn.Dropout(p=dropout_rate),
            nn.ReLU(),
        ]

        block_inner = n_layers * [
            nn.Linear(n_hidden, n_hidden),
            nn.BatchNorm1d(n_hidden),
            nn.ReLU(),
        ]
        return block_inputs + block_inner

    def forward(self, u):
        u_ = u.clone()
        if u_.dim() == 3:
            n_samples, n_cells, n_features = u_.shape
            u0_ = u_[:, :, : self.n_latent].reshape(-1, self.n_latent)
            u_ = u_.reshape(-1, n_features)
            pred_ = self.mod(u_).reshape(n_samples, n_cells, -1)
            scaler_ = self.scaler(u0_).reshape(n_samples, n_cells, -1)
        else:
            pred_ = self.mod(u)
            scaler_ = self.scaler(u[:, : self.n_latent])
        mean = u[..., : self.n_latent] + scaler_ * pred_
        return mean



from copy import deepcopy
import logging
from typing import List, Optional, Union

from anndata import AnnData
import numpy as np
import torch
from tqdm import tqdm
from sklearn.metrics import pairwise_distances

from scvi import REGISTRY_KEYS
from scvi.data import AnnDataManager
from scvi.data.fields import (
    CategoricalJointObsField,
    CategoricalObsField,
    LayerField,
)
from scvi.model.base import UnsupervisedTrainingMixin
from scvi.model.base import BaseModelClass, VAEMixin

from ._module import MrVAE
from ._constants import MRVI_REGISTRY_KEYS

logger = logging.getLogger(__name__)

DEFAULT_TRAIN_KWARGS = dict(
    early_stopping=True,
    early_stopping_patience=15,
    check_val_every_n_epoch=1,
    batch_size=256,
    train_size=0.9,
    plan_kwargs=dict(
        lr=1e-2,
        n_epochs_kl_warmup=20,
    ),
)


class MrVI(UnsupervisedTrainingMixin, VAEMixin, BaseModelClass):
    """
    Multi-resolution Variational Inference (MrVI) :cite:`Boyeau2022mrvi`.

    Parameters
    ----------
    adata
        AnnData object that has been registered via :meth:`~scvi.model.MrVI.setup_anndata`.
    n_latent
        Dimensionality of the latent space.
    n_latent_donor
        Dimensionality of the latent space for sample embeddings.
    linear_decoder_zx
        Whether to use a linear decoder for the decoder from z to x.
    linear_decoder_uz
        Whether to use a linear decoder for the decoder from u to z.
    linear_decoder_uz_scaler
        Whether to incorporate a learned scaler term in the decoder from u to z.
    linear_decoder_uz_scaler_n_hidden
        If `linear_decoder_uz_scaler` is True, the number of hidden
        units in the neural network used to produce the scaler term
        in decoder from u to z.
    px_kwargs
        Keyword args for :class:`~mrvi.components.DecoderZX`.
    pz_kwargs
        Keyword args for :class:`~mrvi.components.DecoderUZ`.
    """

    def __init__(
        self,
        adata,
        **model_kwargs,
    ):
        super().__init__(adata)
        n_cats_per_nuisance_keys = (
            self.adata_manager.get_state_registry(
                MRVI_REGISTRY_KEYS.CATEGORICAL_NUISANCE_KEYS
            ).n_cats_per_key
            if MRVI_REGISTRY_KEYS.CATEGORICAL_NUISANCE_KEYS
            in self.adata_manager.data_registry
            else []
        )

        n_sample = self.summary_stats.n_sample
        n_obs_per_sample = (
            adata.obs.groupby(
                self.adata_manager.get_state_registry(MRVI_REGISTRY_KEYS.SAMPLE_KEY)[
                    "original_key"
                ]
            )
            .size()
            .loc[
                self.adata_manager.get_state_registry(MRVI_REGISTRY_KEYS.SAMPLE_KEY)[
                    "categorical_mapping"
                ]
            ]
            .values
        )
        n_obs_per_sample = torch.from_numpy(n_obs_per_sample).float()
        self.data_splitter = None
        self.module = MrVAE(
            n_input=self.summary_stats.n_vars,
            n_sample=n_sample,
            n_obs_per_sample=n_obs_per_sample,
            n_cats_per_nuisance_keys=n_cats_per_nuisance_keys,
            **model_kwargs,
        )
        self.init_params_ = self._get_init_params(locals())

    @classmethod
    def setup_anndata(
        cls,
        adata: AnnData,
        layer: Optional[str] = None,
        sample_key: Optional[str] = None,
        categorical_nuisance_keys: Optional[List[str]] = None,
        **kwargs,
    ):
        setup_method_args = cls._get_setup_method_args(**locals())
        anndata_fields = [
            LayerField(REGISTRY_KEYS.X_KEY, layer, is_count_data=True),
            CategoricalObsField(MRVI_REGISTRY_KEYS.SAMPLE_KEY, sample_key),
            CategoricalObsField(REGISTRY_KEYS.LABELS_KEY, None),
            CategoricalJointObsField(
                MRVI_REGISTRY_KEYS.CATEGORICAL_NUISANCE_KEYS, categorical_nuisance_keys
            ),
        ]
        adata_manager = AnnDataManager(
            fields=anndata_fields, setup_method_args=setup_method_args
        )
        adata_manager.register_fields(adata, **kwargs)
        cls.register_manager(adata_manager)

    def train(
        self,
        max_epochs: Optional[int] = None,
        use_gpu: Optional[Union[str, int, bool]] = None,
        train_size: float = 0.9,
        validation_size: Optional[float] = None,
        batch_size: int = 128,
        early_stopping: bool = False,
        plan_kwargs: Optional[dict] = None,
        **trainer_kwargs,
    ):
        train_kwargs = dict(
            max_epochs=max_epochs,
            use_gpu=use_gpu,
            train_size=train_size,
            validation_size=validation_size,
            batch_size=batch_size,
            early_stopping=early_stopping,
            **trainer_kwargs,
        )
        train_kwargs = dict(deepcopy(DEFAULT_TRAIN_KWARGS), **train_kwargs)
        plan_kwargs = plan_kwargs or {}
        train_kwargs["plan_kwargs"] = dict(
            deepcopy(DEFAULT_TRAIN_KWARGS["plan_kwargs"]), **plan_kwargs
        )
        super().train(**train_kwargs)

    @torch.no_grad()
    def get_latent_representation(
        self,
        adata: Optional[AnnData] = None,
        indices=None,
        mc_samples: int = 5000,
        batch_size: Optional[int] = None,
        give_z: bool = False,
    ) -> np.ndarray:
        self._check_if_trained(warn=False)
        adata = self._validate_anndata(adata)
        scdl = self._make_data_loader(
            adata=adata, indices=indices, batch_size=batch_size
        )

        u = []
        z = []
        for tensors in tqdm(scdl):
            inference_inputs = self.module._get_inference_input(tensors)
            outputs = self.module.inference(mc_samples=mc_samples, **inference_inputs)
            u.append(outputs["u"].mean(0).cpu())
            z.append(outputs["z"].mean(0).cpu())

        u = torch.cat(u, 0).numpy()
        z = torch.cat(z, 0).numpy()
        return z if give_z else u

    @staticmethod
    def compute_distance_matrix_from_representations(
        representations: np.ndarray, metric: str = "euclidean"
    ) -> np.ndarray:
        """
        Compute distance matrices from counterfactual sample representations.

        Parameters
        ----------
        representations
            Counterfactual sample representations of shape (n_cells, n_sample, n_features).
        metric
            Metric to use for computing distance matrix.
        """
        n_cells, n_donors, _ = representations.shape
        pairwise_dists = np.zeros((n_cells, n_donors, n_donors))
        for i, cell_rep in enumerate(representations):
            d_ = pairwise_distances(cell_rep, metric=metric)
            pairwise_dists[i, :, :] = d_
        return pairwise_dists

    @torch.no_grad()
    def get_local_sample_representation(
        self,
        adata=None,
        batch_size=256,
        mc_samples: int = 10,
        return_distances=False,
    ):
        """
        Computes the local sample representation of the cells in the adata object.

        For each cell, it returns a matrix of size (n_sample, n_features).

        Parameters
        ----------
        adata
            AnnData object to use for computing the local sample representation.
        batch_size
            Batch size to use for computing the local sample representation.
        mc_samples
            Number of Monte Carlo samples to use for computing the local sample representation.
        return_distances
            If ``return_distances`` is ``True``, returns a distance matrix of
            size (n_sample, n_sample) for each cell.
        """
        adata = self.adata if adata is None else adata
        self._check_if_trained(warn=False)
        adata = self._validate_anndata(adata)
        scdl = self._make_data_loader(adata=adata, indices=None, batch_size=batch_size)

        reps = []
        for tensors in tqdm(scdl):
            xs = []
            for sample in range(self.summary_stats.n_sample):
                cf_sample = sample * torch.ones_like(
                    tensors[MRVI_REGISTRY_KEYS.SAMPLE_KEY]
                )
                inference_inputs = self.module._get_inference_input(tensors)
                inference_outputs = self.module.inference(
                    mc_samples=mc_samples, cf_sample=cf_sample, **inference_inputs
                )
                new = inference_outputs["z"]

                xs.append(new[:, :, None])

            xs = torch.cat(xs, 2).mean(0)
            reps.append(xs.cpu().numpy())
        # n_cells, n_sample, n_latent
        reps = np.concatenate(reps, 0)

        if return_distances:
            return self.compute_distance_matrix_from_representations(reps)

        return reps



import torch
import torch.distributions as db
import torch.nn as nn
from scvi import REGISTRY_KEYS
from scvi.module.base import BaseModuleClass, LossRecorder, auto_move_data
from scvi.nn import one_hot
from torch.distributions import kl_divergence as kl

from ._components import DecoderUZ, DecoderZX, LinearDecoderUZ
from ._utils import ConditionalBatchNorm1d, NormalNN
from ._constants import MRVI_REGISTRY_KEYS

DEFAULT_PX_HIDDEN = 32
DEFAULT_PZ_LAYERS = 1
DEFAULT_PZ_HIDDEN = 32


class MrVAE(BaseModuleClass):
    def __init__(
        self,
        n_input,
        n_sample,
        n_obs_per_sample,
        n_cats_per_nuisance_keys,
        n_latent=10,
        n_latent_sample=2,
        linear_decoder_zx=True,
        linear_decoder_uz=True,
        linear_decoder_uz_scaler=False,
        linear_decoder_uz_scaler_n_hidden=32,
        px_kwargs=None,
        pz_kwargs=None,
    ):
        super().__init__()
        px_kwargs = dict(n_hidden=DEFAULT_PX_HIDDEN)
        if px_kwargs is not None:
            px_kwargs.update(px_kwargs)
        pz_kwargs = dict(n_layers=DEFAULT_PZ_LAYERS, n_hidden=DEFAULT_PZ_HIDDEN)
        if pz_kwargs is not None:
            pz_kwargs.update(pz_kwargs)

        self.n_cats_per_nuisance_keys = n_cats_per_nuisance_keys
        self.n_sample = n_sample
        assert n_latent_sample != 0
        self.sample_embeddings = nn.Embedding(n_sample, n_latent_sample)

        n_nuisance = sum(self.n_cats_per_nuisance_keys)
        # Generative model
        self.px = DecoderZX(
            n_latent + n_nuisance,
            n_input,
            n_nuisance=n_nuisance,
            linear_decoder=linear_decoder_zx,
            **px_kwargs,
        )
        self.qu = NormalNN(128 + n_latent_sample, n_latent, n_categories=1)
        self.ql = NormalNN(n_input, 1, n_categories=1)

        self.linear_decoder_uz = linear_decoder_uz
        if linear_decoder_uz:
            self.pz = LinearDecoderUZ(
                n_latent,
                self.n_sample,
                n_latent,
                scaler=linear_decoder_uz_scaler,
                scaler_n_hidden=linear_decoder_uz_scaler_n_hidden,
            )
        else:
            self.pz = DecoderUZ(
                n_latent,
                n_latent_sample,
                n_latent,
                **pz_kwargs,
            )
        self.n_obs_per_sample = nn.Parameter(n_obs_per_sample, requires_grad=False)

        self.x_featurizer = nn.Sequential(nn.Linear(n_input, 128), nn.ReLU())
        self.bnn = ConditionalBatchNorm1d(128, n_sample)
        self.x_featurizer2 = nn.Sequential(nn.Linear(128, 128), nn.ReLU())
        self.bnn2 = ConditionalBatchNorm1d(128, n_sample)

    def _get_inference_input(self, tensors, **kwargs):
        x = tensors[REGISTRY_KEYS.X_KEY]
        sample_index = tensors[MRVI_REGISTRY_KEYS.SAMPLE_KEY]
        categorical_nuisance_keys = tensors[
            MRVI_REGISTRY_KEYS.CATEGORICAL_NUISANCE_KEYS
        ]
        return dict(
            x=x,
            sample_index=sample_index,
            categorical_nuisance_keys=categorical_nuisance_keys,
        )

    @auto_move_data
    def inference(
        self,
        x,
        sample_index,
        categorical_nuisance_keys,
        mc_samples=1,
        cf_sample=None,
        use_mean=False,
    ):
        x_ = torch.log1p(x)

        sample_index_cf = sample_index if cf_sample is None else cf_sample
        zsample = self.sample_embeddings(sample_index_cf.long().squeeze(-1))
        zsample_ = zsample
        if mc_samples >= 2:
            zsample_ = zsample[None].expand(mc_samples, *zsample.shape)

        nuisance_oh = []
        for dim in range(categorical_nuisance_keys.shape[-1]):
            nuisance_oh.append(
                one_hot(
                    categorical_nuisance_keys[:, [dim]],
                    self.n_cats_per_nuisance_keys[dim],
                )
            )
        nuisance_oh = torch.cat(nuisance_oh, dim=-1)

        x_feat = self.x_featurizer(x_)
        x_feat = self.bnn(x_feat, sample_index)
        x_feat = self.x_featurizer2(x_feat)
        x_feat = self.bnn2(x_feat, sample_index)
        if x_.ndim != zsample_.ndim:
            x_feat_ = x_feat[None].expand(mc_samples, *x_feat.shape)
            nuisance_oh = nuisance_oh[None].expand(mc_samples, *nuisance_oh.shape)
        else:
            x_feat_ = x_feat

        inputs = torch.cat([x_feat_, zsample_], -1)
        # inputs = x_feat_
        qu = self.qu(inputs)
        if use_mean:
            u = qu.loc
        else:
            u = qu.rsample()

        if self.linear_decoder_uz:
            z = self.pz(u, sample_index_cf)
        else:
            inputs = torch.cat([u, zsample_], -1)
            z = self.pz(inputs)
        library = torch.log(x.sum(1)).unsqueeze(1)

        return dict(
            qu=qu,
            u=u,
            z=z,
            zsample=zsample,
            library=library,
            nuisance_oh=nuisance_oh,
        )

    def get_z(self, u, zsample=None, sample_index=None):
        if sample_index is not None:
            zsample = self.sample_embeddings(sample_index.long().squeeze(-1))
            zsample = zsample
        else:
            zsample_ = zsample
        inputs = torch.cat([u, zsample_], -1)
        z = self.pz(inputs)
        return z

    def _get_generative_input(self, tensors, inference_outputs, **kwargs):
        res = dict(
            z=inference_outputs["z"],
            library=inference_outputs["library"],
            nuisance_oh=inference_outputs["nuisance_oh"],
        )

        return res

    @auto_move_data
    def generative(
        self,
        z,
        library,
        nuisance_oh,
    ):

        inputs = torch.concat([z, nuisance_oh], dim=-1)
        px = self.px(inputs, size_factor=library.exp())
        h = px.mu / library.exp()

        pu = db.Normal(0, 1)
        return dict(px=px, pu=pu, h=h)

    def loss(
        self,
        tensors,
        inference_outputs,
        generative_outputs,
        kl_weight: float = 1.0,
    ):

        reconstruction_loss = (
            -generative_outputs["px"].log_prob(tensors[REGISTRY_KEYS.X_KEY]).sum(-1)
        )
        kl_u = kl(inference_outputs["qu"], generative_outputs["pu"]).sum(-1)
        kl_local_for_warmup = kl_u

        weighted_kl_local = kl_weight * kl_local_for_warmup
        loss = torch.mean(reconstruction_loss + weighted_kl_local)

        kl_local = torch.tensor(0.0)
        kl_global = torch.tensor(0.0)
        return LossRecorder(
            loss,
            reconstruction_loss,
            kl_local,
            kl_global,
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mrvi']

package_data = \
{'': ['*']}

install_requires = \
['anndata>=0.7.5', 'scvi-tools>=0.17.0']

extras_require = \
{':(python_version < "3.8") and (extra == "docs")': ['typing_extensions'],
 ':python_version < "3.8"': ['importlib-metadata>=1.0,<2.0'],
 'dev': ['black>=20.8b1',
         'codecov>=2.0.8',
         'flake8>=3.7.7',
         'isort>=5.7',
         'jupyter>=1.0',
         'loompy>=3.0.6',
         'nbconvert>=5.4.0',
         'nbformat>=4.4.0',
         'pre-commit>=2.7.1',
         'pytest>=4.4',
         'scanpy>=1.6'],
 'docs': ['ipython>=7.1.1',
          'nbsphinx',
          'nbsphinx-link',
          'pydata-sphinx-theme>=0.4.0',
          'scanpydoc>=0.5',
          'sphinx>=4.1,<4.4',
          'sphinx-autodoc-typehints',
          'sphinx-rtd-theme']}

setup_kwargs = {
    'name': 'mrvi',
    'version': '0.1.1',
    'description': 'Multi-resolution analysis of single-cell data.',
    'long_description': '# Multi-resolution Variational Inference\n\nMulti-resolution Variational Inference (MrVI) is a package for analysis of sample-level heterogeneity in multi-site, multi-sample single-cell omics data. Built with [scvi-tools](https://scvi-tools.org).\n\n---\n\nTo install, run:\n\n```\npip install mrvi\n```\n\n`mrvi.MrVI` follows the same API used in scvi-tools.\n\n```python\nimport mrvi\nimport anndata\n\nadata = anndata.read_h5ad("path/to/adata.h5ad")\n#\xa0Sample (e.g. donors, perturbations, etc.) should go in sample_key\n# Sites, plates, and other factors should go in categorical_nuisance_keys\nmrvi.MrVI.setup_anndata(adata, sample_key="donor", categorical_nuisance_keys=["site"])\nmrvi_model = mrvi.MrVI(adata)\nmrvi_model.train()\n# Get z representation\nadata.obsm["X_mrvi_z"] = mrvi_model.get_latent_representation(give_z=True)\n# Get u representation\nadata.obsm["X_mrvi_u"] = mrvi_model.get_latent_representation(give_z=False)\n# Cells by n_sample by n_latent\ncell_sample_representations = mrvi_model.get_local_sample_representation()\n# Cells by n_sample by n_sample\ncell_sample_sample_distances = mrvi_model.get_local_sample_representation(return_distances=True)\n```\n\n## Citation\n\n```\n@article {Boyeau2022.10.04.510898,\n\tauthor = {Boyeau, Pierre and Hong, Justin and Gayoso, Adam and Jordan, Michael and Azizi, Elham and Yosef, Nir},\n\ttitle = {Deep generative modeling for quantifying sample-level heterogeneity in single-cell omics},\n\telocation-id = {2022.10.04.510898},\n\tyear = {2022},\n\tdoi = {10.1101/2022.10.04.510898},\n\tpublisher = {Cold Spring Harbor Laboratory},\n\tabstract = {Contemporary single-cell omics technologies have enabled complex experimental designs incorporating hundreds of samples accompanied by detailed information on sample-level conditions. Current approaches for analyzing condition-level heterogeneity in these experiments often rely on a simplification of the data such as an aggregation at the cell-type or cell-state-neighborhood level. Here we present MrVI, a deep generative model that provides sample-sample comparisons at a single-cell resolution, permitting the discovery of subtle sample-specific effects across cell populations. Additionally, the output of MrVI can be used to quantify the association between sample-level metadata and cell state variation. We benchmarked MrVI against conventional meta-analysis procedures on two synthetic datasets and one real dataset with a well-controlled experimental structure. This work introduces a novel approach to understanding sample-level heterogeneity while leveraging the full resolution of single-cell sequencing data.Competing Interest StatementN.Y. is an advisor and/or has equity in Cellarity, Celsius Therapeutics, and Rheos Medicine.},\n\tURL = {https://www.biorxiv.org/content/early/2022/10/06/2022.10.04.510898},\n\teprint = {https://www.biorxiv.org/content/early/2022/10/06/2022.10.04.510898.full.pdf},\n\tjournal = {bioRxiv}\n}\n```\n',
    'author': 'Pierre Boyeau',
    'author_email': 'pierreboyeau@berkeley.edu',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/YosefLab/mrvi',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7.2,<4.0',
}


setup(**setup_kwargs)



from typing import NamedTuple


class _MRVI_REGISTRY_KEYS_NT(NamedTuple):
    SAMPLE_KEY: str = "sample"
    CATEGORICAL_NUISANCE_KEYS: str = "categorical_nuisance_keys"


MRVI_REGISTRY_KEYS = _MRVI_REGISTRY_KEYS_NT()
